from typing import List, Optional
from unicodedata import normalize
from dataclasses import asdict

from py_ecc.optimized_bls12_381 import curve_order as bls_curve_order
from py_ecc.bls import G2ProofOfPossession as bls

from geodefi.utils.bls.serialize import (
    DepositMessage,
    compute_deposit_domain,
    compute_signing_root,
    DepositData,
)

from geodefi.utils.bls.keystore import ScryptKeystore, Keystore

from geodefi.utils.bls.crypto import SHA256, HKDF, _HKDF_mod_r, PBKDF2

from geodefi.globals.constants import (
    DEPOSIT_CLI_VERSION,
    WORD_LIST_PATH,
    ETH1_ADDRESS_WITHDRAWAL_PREFIX,
)
from geodefi.globals.beacon import (
    DEPOSIT_SIZE,
    GENESIS_FORK_VERSION,
    DEPOSIT_NETWORK_NAME,
)


import os
from random import SystemRandom

_sysrand = SystemRandom()

randbits = _sysrand.getrandbits


def from_mnemonic(
    mnemonic: str,
    mnemonic_password: str,
    chain_id: int,
    start_index: int,
    num_keys: int,
    hex_eth1_withdrawal_address,
):
    key_indices = range(start_index, start_index + num_keys)

    # Set path as EIP-2334 format
    # https://eips.ethereum.org/EIPS/eip-2334
    purpose = "12381"
    coin_type = "3600"

    deposit_data = []
    keystore_data = []

    for index in key_indices:
        for_each_pubkey = []
        for amount in [DEPOSIT_SIZE.PROPOSAL.value, DEPOSIT_SIZE.STAKE.value]:
            account = str(index)
            withdrawal_key_path = f"m/{purpose}/{coin_type}/{account}/0"
            signing_key_path = f"{withdrawal_key_path}/0"

            withdrawal_sk = mnemonic_and_path_to_key(
                mnemonic=mnemonic,
                path=withdrawal_key_path,
                password=mnemonic_password,
            )
            signing_sk = mnemonic_and_path_to_key(
                mnemonic=mnemonic,
                path=signing_key_path,
                password=mnemonic_password,
            )
            """
            Return a single deposit datum for 1 validator including all
            the information needed to verify and process the deposit.
            """

            withdrawal_pk = bls.SkToPk(withdrawal_sk)
            withdrawal_credentials = SHA256(withdrawal_pk)[1:]

            if hex_eth1_withdrawal_address is None:
                raise Exception("#TODO")

            else:
                withdrawal_credentials = ETH1_ADDRESS_WITHDRAWAL_PREFIX
                withdrawal_credentials += b"\x00" * 11
                withdrawal_credentials += hex_eth1_withdrawal_address

            deposit_message = DepositMessage(
                pubkey=bls.SkToPk(signing_sk),
                withdrawal_credentials=withdrawal_credentials,
                amount=amount,
            )

            domain = compute_deposit_domain(
                fork_version=GENESIS_FORK_VERSION[chain_id]
            )
            signing_root = compute_signing_root(deposit_message, domain)
            signed_deposit = DepositData(
                **deposit_message.as_dict(),
                signature=bls.Sign(signing_sk, signing_root),
            )

            signed_deposit_datum = signed_deposit
            datum_dict = signed_deposit_datum.as_dict()
            datum_dict.update(
                {"deposit_message_root": deposit_message.hash_tree_root}
            )
            datum_dict.update(
                {"deposit_data_root": signed_deposit_datum.hash_tree_root}
            )
            datum_dict.update({"fork_version": GENESIS_FORK_VERSION[chain_id]})
            datum_dict.update({"network_name": DEPOSIT_NETWORK_NAME[chain_id]})
            datum_dict.update({"deposit_cli_version": DEPOSIT_CLI_VERSION})

            for_each_pubkey.append(datum_dict)

        keystore_data.append(
            signing_keystore(signing_sk, signing_key_path, mnemonic_password)
        )
        deposit_data.append(for_each_pubkey)

    return deposit_data, keystore_data


def get_mnemonic(*, entropy: Optional[bytes] = None) -> str:
    """
    Return a mnemonic string in a given `language` based on `entropy` via the calculated checksum.

    Ref: https://github.com/bitcoin/bips/blob/master/bip-0039.mediawiki#generating-the-mnemonic
    """
    if entropy is None:
        entropy = randbits(256).to_bytes(32, "big")
    entropy_length = len(entropy) * 8
    checksum_length = entropy_length // 32
    checksum = _get_checksum(entropy)
    entropy_bits = int.from_bytes(entropy, "big") << checksum_length
    entropy_bits += checksum
    entropy_length += checksum_length
    mnemonic = []
    word_list = _get_word_list(WORD_LIST_PATH)
    for i in range(entropy_length // 11 - 1, -1, -1):
        index = (entropy_bits >> i * 11) & 2**11 - 1
        word = _index_to_word(word_list, index)
        mnemonic.append(word)
    return " ".join(mnemonic)


def _get_checksum(entropy: bytes) -> int:
    """
    Determine the index of the checksum word given the entropy
    """
    _validate_entropy_length(entropy)
    checksum_length = len(entropy) // 4
    return int.from_bytes(SHA256(entropy), "big") >> (256 - checksum_length)


def _get_word_list(path: str) -> List[str]:
    """
    Given the language and path to the wordlist, return the list of BIP39 words.

    Ref: https://github.com/bitcoin/bips/blob/master/bip-0039/bip-0039-wordlists.md
    """
    dirty_list = open(
        os.path.join(path, "word_list.txt"), encoding="utf-8"
    ).readlines()
    return [word.replace("\n", "") for word in dirty_list]


def _index_to_word(word_list: List[str], index: int) -> str:
    """
    Return the corresponding word for the supplied index while stripping out '\\n' chars.
    """
    if index >= 2048:
        raise IndexError(f"`index` should be less than 2048. Got {index}.")
    return word_list[index]


def _validate_entropy_length(entropy: bytes) -> None:
    entropy_length = len(entropy) * 8
    if entropy_length not in range(128, 257, 32):
        raise IndexError(
            f"`entropy_length` should be in [128, 160, 192, 224, 256]. Got {entropy_length}."
        )


def path_to_nodes(path: str) -> List[int]:
    """
    Maps from a path string to a list of indices where each index represents the corresponding level in the path.
    """
    path = path.replace(" ", "")
    if not set(path).issubset(set("m1234567890/")):
        raise ValueError(f"Invalid path {path}")

    indices = path.split("/")

    if indices[0] != "m":
        raise ValueError(
            f"The first character of path should be `m`. Got {indices[0]}."
        )
    indices.pop(0)

    return [int(index) for index in indices]


def mnemonic_and_path_to_key(*, mnemonic: str, path: str, password: str) -> int:
    """
    Return the SK at position `path`, derived from `mnemonic`. The password is to be
    compliant with BIP39 mnemonic that use passwords, but is not used by this CLI outside of tests.
    """
    seed = get_seed(mnemonic=mnemonic, password=password)
    sk = derive_master_SK(seed)
    for node in path_to_nodes(path):
        sk = derive_child_SK(parent_SK=sk, index=node)
    return sk


def get_seed(*, mnemonic: str, password: str) -> bytes:
    """
    Derive the seed for the pre-image root of the tree.

    Ref: https://github.com/bitcoin/bips/blob/master/bip-0039.mediawiki#from-mnemonic-to-seed
    """
    encoded_mnemonic = normalize("NFKD", mnemonic).encode("utf-8")
    salt = normalize("NFKD", "mnemonic" + password).encode("utf-8")
    return PBKDF2(
        password=encoded_mnemonic, salt=salt, dklen=64, c=2048, prf="sha512"
    )


def derive_child_SK(*, parent_SK: int, index: int) -> int:
    """
    Given a parent SK `parent_SK`, return the child SK at the supplied `index`.

    Ref: https://github.com/ethereum/EIPs/blob/master/EIPS/eip-2333.md#derive_child_sk
    """
    if index < 0 or index >= 2**32:
        raise IndexError(
            f"`index` should be greater than or equal to  0 and less than 2**32. Got index={index}."
        )
    lamport_PK = _parent_SK_to_lamport_PK(parent_SK=parent_SK, index=index)
    return _HKDF_mod_r(IKM=lamport_PK)


def derive_master_SK(seed: bytes) -> int:
    """
    Given a seed, derive the master SK.

    Ref: https://github.com/ethereum/EIPs/blob/master/EIPS/eip-2333.md#derive_master_sk
    """
    if len(seed) < 32:
        raise ValueError(
            f"`len(seed)` should be greater than or equal to 32. Got {len(seed)}."
        )
    return _HKDF_mod_r(IKM=seed)


def _flip_bits_256(input: int) -> int:
    """
    Flips 256 bits worth of `input`.
    """
    return input ^ (2**256 - 1)


def _IKM_to_lamport_SK(*, IKM: bytes, salt: bytes) -> List[bytes]:
    """
    Derives the lamport SK for a given `IKM` and `salt`.

    Ref: https://github.com/ethereum/EIPs/blob/master/EIPS/eip-2333.md#ikm_to_lamport_sk
    """
    OKM = HKDF(salt=salt, IKM=IKM, L=8160)
    lamport_SK = [OKM[i : i + 32] for i in range(0, 8160, 32)]
    return lamport_SK


def _parent_SK_to_lamport_PK(*, parent_SK: int, index: int) -> bytes:
    """
    Derives the `index`th child's lamport PK from the `parent_SK`.

    Ref: https://github.com/ethereum/EIPs/blob/master/EIPS/eip-2333.md#parent_sk_to_lamport_pk
    """
    salt = index.to_bytes(4, byteorder="big")
    IKM = parent_SK.to_bytes(32, byteorder="big")
    lamport_0 = _IKM_to_lamport_SK(IKM=IKM, salt=salt)
    not_IKM = _flip_bits_256(parent_SK).to_bytes(32, byteorder="big")
    lamport_1 = _IKM_to_lamport_SK(IKM=not_IKM, salt=salt)
    lamport_SKs = lamport_0 + lamport_1
    lamport_PKs = [SHA256(sk) for sk in lamport_SKs]
    compressed_PK = SHA256(b"".join(lamport_PKs))
    return compressed_PK


def signing_keystore(
    signing_sk: int, signing_key_path: str, password: str
) -> Keystore:
    secret = signing_sk.to_bytes(32, "big")
    return asdict(
        ScryptKeystore.encrypt(
            secret=secret, password=password, path=signing_key_path
        )
    )



from dotenv import dotenv_values
import sys
import os
import pytest
import re

from geodefi.utils.merkle import StandartMerkleTree
from geodefi import Geode
from geodefi.classes.validator import Validator


sys.path.append(os.getcwd())


env = dotenv_values(".env")


G = Geode(exec_api=env["EXECUTION_API"], cons_key=env["CONSENSUS_KEY"])
PORTAL = G.Portal

POOL_LIST = [
    50016835115526216130031110555486827201953559012021267556883950029143900999178,
    29228457249232120346521013786824808088246537603535847808963148138747123868265,
]

OPERATOR_LIST = [
    114391297015478800753082638170652680401082080549997516459063441314156612391510,
    106440257855610140438147021303249733137151692982873052796064618175630395030085,
]


def test_init():

    myPool = PORTAL.pool(POOL_LIST[0])
    myVal = myPool.validators(0)

    assert isinstance(myVal, Validator)

    # '0x9326f6c07f8abd082ef82b19279cbba7616b0395fb947d50cd2d5fef303dd613abe31087077a67faa477c0631cc7228d'
    assert myVal.pubkey == "0x"
    assert myVal.w3 == G.w3
    assert myVal.network == 5
    assert myVal.beacon == G.Beacon
    assert myVal.portal.address == G.Portal.address


# Portal


myPool = PORTAL.pool(POOL_LIST[0])
myVal = myPool.validators(0)


def test_state():
    assert myVal.state == 0


def test_index():
    assert myVal.index == 0


def test_poolId():
    assert myVal.poolId == 0


def test_operatorId():
    assert myVal.operatorId == 0


def test_poolFee():
    assert myVal.poolFee == 0


def test_operatorFee():
    assert myVal.operatorFee == 0


def test_earlyExitFee():
    assert myVal.earlyExitFee == 0


def test_createdAt():
    assert myVal.createdAt == 0


def test_expectedExit():
    assert myVal.expectedExit == 0


def test_signature31():

    # == b'\x94\xc0\x18~I\x0e\xc3\x96r&\xd3\xc3\xce\xbc\xf0\xb0t\xbf\xa0Iq\xe5+\x95t\x8e\x91\x93?\x93\xfc?\x93g}\x94tM\xf5 \x89|\x99\xd3sn\xd1\xdb\x08\xa8!i\x813\xc2b\xb3SdB\x95Y\xa1\xb0z\xc4\x85`\xd2z.g\x88Dq\xf8R/g\xae\nB\xfa\xaa\xee!~\x9c@\xe0\\\xd91(\xad\xdb'
    assert isinstance(myVal.signature31, bytes)
    # == '94c0187e490ec3967226d3c3cebcf0b074bfa04971e52b95748e91933f93fc3f93677d94744df520897c99d3736ed1db08a821698133c262b35364429559a1b07ac48560d27a2e67884471f8522f67ae0a42faaaee217e9c40e05cd93128addb'
    # assert int(myVal.signature31.hex(), 16)


def test_str():
    # 9326f6c07f8abd082ef82b19279cbba7616b0395fb947d50cd2d5fef303dd613abe31087077a67faa477c0631cc7228d'
    assert str(myVal) == "Validator Object: 0x"


# BEACON


"""
def test_activationEligibilityEpoch():
    assert myVal.activationEligibilityEpoch == 0


def test_activationEpoch():
    assert myVal.activationEpoch == 0


def test_balance():
    assert myVal.balance == 0


def test_effectiveBalance():
    assert myVal.effectiveBalance == 0


def test_exitepoch():
    assert myVal.exitepoch == 0


def test_lastAttestationSlot():
    assert myVal.lastAttestationSlot == 0


def test_slashed():
    assert myVal.slashed == 0


def test_status():
    assert myVal.status == 0


def test_validatorIndex():
    assert myVal.validatorIndex == 0


def test_withdrawableEpoch():
    assert myVal.withdrawableEpoch == 0


def test_withdrawalCredentials():
    assert myVal.withdrawalCredentials == 0


def test_totalWithdrawals():
    assert myVal.totalWithdrawals == 0
"""



from math import pow, ceil, log2
from functools import cache
import typing as t
from eth_abi import encode
from Crypto.Hash import keccak as _keccak


def keccak256(x: bytes) -> bytes:

    if isinstance(x, str):
        x = bytes.fromhex(x[2:])

    k = _keccak.new(digest_bits=256)
    k.update(x)

    return k.hexdigest()


class Tree:

    def __init__(self, types, values, tree='') -> None:
        self._types: t.List = types
        self._values: t.List = values
        self._tree: t.List = tree
        self._format = 'standard-v1'
        self._root = tree[1]

    def __repr__(self):
        _tree_repr = "'\n\t  '".join(self._tree_hexify(self._tree[1:]))

        return f"MerkleTree(\n\tformat: '{self._format}',\n\ttree:\n\t  '{_tree_repr}',\n\tvalues:\n\t\t{self._values},\n\troot:\n\t  {self.root}\n\tleaf_encoding:  {self._types})"

    @property
    def tree(self):
        return self._tree[1:]

    @property
    def root(self):
        return '0x' + self._tree[1]

    @property
    def values(self):
        return self._values

    @property
    def format(self):
        return self._format

    @property
    def leaf_encodings(self):
        return self._types

    def _tree_hexify(self, tree):
        t = []
        for i in tree:
            t.append('0x'+i)
        return t


class StandartMerkleTree:

    def __init__(self, sort=True):
        """
        The merkle tree works like a heap array.
        In the worst case scenario, there is O(nlogn) memory complexity,
        but heap array is preferred because it can be implemented more easily.
        It only works with keccak256 and is compatible with Solidity.

        param:sort == Two child node will be sorted when concatting to get hash of parent Default:True
        """
        self._tree = []
        self._generated = False

        self._depth = 0  # 0 or 1 nodes => 0 depth, 2 nodes => 1 depth, 4 nodes => 2 depth,
        self._numberOfNodes = 0  # The lenght of leaves

        # Always power of 2, How many memory should be allocated.
        self._sizeOfTree = 0
        self._sort = sort  # Recommentation: Always use sort = True for solidity compatilibty

    def leafToHash(self, leaf, types):
        return keccak256(bytes.fromhex(keccak256(encode(types, leaf))))

    def of(self, values: t.List, types: t.List):
        """
        Hash every node with keccak until the root has been calculated.
        Always first index of the tree is None (tree[0] = None)
        If the nodes is not power of 2, None type nodes will be generated.
        Some parents nodes will be None also.

        Tree Structure:
                Root:    tree[1]
                        /      |
                     tree[2]    tree[3]
                    /      |    /      |
             tree[4] tree[5]  tree[6]  tree[7]
                        ...

        How to handle None types (Notice (tree[6] == tree[3]))

                Root:    H( H(0xaa.., 0xbb..) , 0xcc..)
                        /                   |
                H(0xaa.., 0xbb..)           0xccc...
                    /      |                /       |
              0xaaa.. 0xbbb..          0xccc..      None

        param:keccakTheLeaves == If true: the leaves (nodes) will also be hashed.
        Otherwise, tree assumes you submit the hex-string which result of some hashes as leaves
        Default: True
        """

        assert len(values) != 0, "Values cannot be empty."
        assert len(values[0]) == len(
            types), "Number of values and types must be equal"

        # Create space for tree.

        numberOfNodes = len(values)  # Lets assume numberOfNodes is 13,
        # Then, the tree size will be 32, so depth will be 5
        depth = 1 + ceil(log2(numberOfNodes))
        sizeOfTree = int(pow(2, depth))

        self._tree = [None] * sizeOfTree
        self._sizeOfTree = sizeOfTree
        self._depth = depth
        self._numberOfNodes = numberOfNodes

        startIndex = sizeOfTree//2
        base_leaves = []
        # Take all values like ['123',True,'0xabcd']
        for v in values:

            for i, type in enumerate(types):
                # type check
                if type in ['bytes', 'bytes32']:
                    # Hex string
                    if isinstance(v[i], str):
                        if v[i][:2] == '0x':
                            try:
                                int(v[i][2:], 16)
                            except ValueError:
                                raise ValueError(f"{v[i]} is not hex-string.")

                            v[i] = bytes.fromhex(v[i][2:])

                        else:
                            try:
                                int(v[i], 16)
                            except ValueError:
                                raise ValueError(f"{v[i]} is not hex-string.")
                            v[i] = bytes.fromhex(v[i])

                    elif isinstance(v[i], bytes):
                        pass
                    else:
                        raise ValueError(f"{v[i]} is not bytes.")

                elif type in ['int', 'uint', 'uint8', 'uint16', 'uint32', 'uint64', 'uint128', 'uint256', 'int8', 'int16', 'int32', 'int64', 'int128', 'int256']:
                    try:
                        v[i] = int(v[i])
                    except ValueError:
                        raise ValueError(
                            f"{v[i]} is not integer or unsigned integer.")
                elif 'bool' == type:

                    if isinstance(v[i], bool):
                        pass
                    elif v[i] in ['true', 'True']:
                        v[i] = True
                    elif v[i] in ['False', 'false']:
                        v[i] = False
                    else:
                        raise ValueError(f"{v[i]} is not boolean.")

            leaf = encode(types, v)

            base_leaves.append(keccak256(bytes.fromhex(keccak256(leaf))))

        for i, leaf in enumerate(base_leaves):
            self._tree[i + startIndex] = leaf

        # Generate Upper Layers
        while startIndex >= 1:
            startIndex //= 2  # Traverse like ...32 - 16 - 8 - 4 - 2 - 1

            for i in range(startIndex):
                left = self._tree[(i + startIndex)*2]
                right = self._tree[(i + startIndex)*2 + 1]

                if (right is None and left is None):
                    self._tree[i + startIndex] = None
                elif (right is None):
                    self._tree[i + startIndex] = left
                else:
                    self._tree[i + startIndex] = keccak256(
                        self.concat(left, right, self._sort))

        self._generated = True

        return Tree(types=types, values=values, tree=self._tree)

    def getProofs(self, leaf):
        """
        Tree Structure:
                Root:    tree[1]
                        /      |
                     tree[2]    tree[3]
                    /      |    /      |
             tree[4] tree[5]  tree[6]  tree[7]
                        ...

        Sample:
                leaf = tree[4]
                returns = [tree[5], tree[3]]

        If you generate proofs by tree[4],,
            tree[5] and tree[3] can re-generate the root like
            H(tree[4], tree[5]) (which is equal to tree[2])

            and
            H(tree[2], tree[3]) can re-generate the tree[1] which is ROOT


        :param: leaf => get some node hashes which can re-generate the root
                If leaf is not in the tree raise Exception
        """

        proofList = []
        indexOfNode = 0

        # Search within the right half of the tree
        for i in range(self.sizeOfTree//2, self.sizeOfTree):
            if (self._tree[i] == leaf):
                indexOfNode = i
                break

        # If you could not find the leaf
        if indexOfNode == 0:
            raise Exception(f'{leaf} is not a leaf')

        while True:
            # Get sibling of the node
            # Sibling == If odd, substract 1, if even, add 1.
            # Sample siblings: (4,5), (45,44), (1245125, 1245124)
            proof = self._tree[
                indexOfNode + 1] if i % 2 == 0 else self._tree[
                    indexOfNode - 1]

            proofList.append('0x'+proof)
            # Divide 2 means go to parent node
            indexOfNode //= 2

            if (indexOfNode == 1):
                break

        return proofList

    # def verifyProof(self, leaf, proofs) -> bool:
    #    # To be Implemented
    #    pass

    def printTree(self):
        if self.generated:
            print("============== TREE ==============")

            self.nextNode(1)

    @cache
    def nextNode(self, i):
        if i > self._sizeOfTree - 1:
            return

        depthOfIndex = int(log2(i)) + 1

        if self._tree[i] is not None:
            print(i, '|  ' * depthOfIndex + '|'+'_' *
                  depthOfIndex + ': ' + self._tree[i])

        # RECURSION
        self.nextNode(i*2)
        self.nextNode(i*2 + 1)

    @staticmethod
    def concat(a: str, b: str, sort=True) -> str:
        """
        concat two strings like

        0xAAA + 0xBBB => 0xAAABBB

        param a: first string
        param b: first string
        param sort: if True, sort the strings before concatting, Default: True
        """

        if not isinstance(a, str) or not isinstance(b, str):
            raise ValueError("concat function works with strings")

        if len(a) > 2 and a[:2] == '0x':
            a = a.lstrip('0x')

        if len(b) > 2 and b[:2] == '0x':
            b = b.lstrip('0x')

        if sort:
            return ('0x' + a + b) if a < b else ('0x' + b + a)
        else:
            return ('0x' + a + b)

    def __str__(self):
        return f"The tree has no leaves yet." if not self._generated else f"Tree Root Hash: {self.root}"

    def reset(self):
        """
        Clear the list.
        """
        self._tree = []
        self._generated = False
        self._depth = 0
        self._sizeOfTree = 0
        self._numberOfNodes = 0

    # Properties

    @property
    def root(self):
        if not self._generated:
            # Tree is not generated yet.
            return None
        return self._tree[1]

    @property
    def generated(self):
        return self._generated

    @property
    def numberOfNodes(self):
        return self._numberOfNodes

    @property
    def tree(self):
        return self._tree[1:]

    @property
    def sizeOfTree(self):
        return self._sizeOfTree

    @property
    def depth(self):
        return self._depth

    @property
    def leaves(self):
        return self._tree[-self.numberOfNodes:]



from dotenv import dotenv_values
import pytest

from geodefi.utils.bls.validate import (
    SHA256,
    validate_deposit,
    validate_parameters,
)
from geodefi.utils.bls.serialize import (
    DepositData,
    DepositMessage,
    compute_signing_root,
    compute_deposit_fork_data_root,
    compute_deposit_domain,
)
from geodefi.globals.constants import (
    DOMAIN_DEPOSIT,
    ZERO_BYTES32,
)

from eth_typing import (
    BLSPubkey,
    BLSSignature,
)

from py_ecc.bls import G2ProofOfPossession as bls

"""
def test_validate_parameters():
    pass


def test_validate_deposit():
    pass


def test_compute_deposit_domain():
    pass


def test_compute_deposit_fork_data_root():
    pass
"""
"""


def test_compute_deposit_fork_data_root_valid():
    # Test with a valid current_version (4 bytes)
    current_version = b'\x01\x23\x45\x67'
    expected_root = b'\x01\x23\x45\x67' + \
        ZERO_BYTES32[4:]  # Hash tree root of ForkData

    assert compute_deposit_fork_data_root(
        current_version) == b'\xe6"\xc6~\x82\xe5\xe8\x05K@\xf2\x82\x93\x871\xa1\xc0\x06\xe0\x1eZd#\xa6|\x1f\xc7\xce\xb1\xdb\\G'



def test_compute_deposit_fork_data_root_invalid():
    # Test with an invalid current_version (not 4 bytes)
    invalid_version = b'\x01\x23\x45'
    with pytest.raises(ValueError, match="Fork version should be in 4 bytes. Got 3."):
        compute_deposit_fork_data_root(invalid_version)




def test_compute_deposit_fork_data_root_zero_bytes():
    # Test with all zeros in current_version
    current_version = ZERO_BYTES32[:4]
    expected_root = ZERO_BYTES32

    assert compute_deposit_fork_data_root(
        current_version) == b"\xf5\xa5\xfdB\xd1j 0'\x98\xefn\xd3\t\x97\x9bC\x00=# \xd9\xf0\xe8\xea\x981\xa9'Y\xfbK"

"""

# Helper function to create a valid DepositMessage for testing



from Crypto.Cipher import AES as _AES

from Crypto.Hash import (
    SHA256 as _sha256,
    SHA512 as _sha512,
)
from Crypto.Protocol.KDF import (
    scrypt as _scrypt,
    HKDF as _HKDF,
    PBKDF2 as _PBKDF2,
)

from py_ecc.optimized_bls12_381 import curve_order as bls_curve_order


def AES_128_CTR(*, key: bytes, iv: bytes):
    if len(key) != 16:
        raise ValueError(f"The key length should be 16. Got {len(key)}.")
    return _AES.new(key=key, mode=_AES.MODE_CTR, initial_value=iv, nonce=b"")


def SHA256(x: bytes) -> bytes:
    return _sha256.new(x).digest()


def HKDF(*, salt: bytes, IKM: bytes, L: int, info: bytes = b"") -> bytes:
    res = _HKDF(master=IKM, key_len=L, salt=salt, hashmod=_sha256, context=info)
    # PyCryptodome can return Tuple[bytes]
    return res if isinstance(res, bytes) else res[0]


def _HKDF_mod_r(*, IKM: bytes, key_info: bytes = b"") -> int:
    """
    Hashes the IKM using HKDF and returns the answer as an int modulo r, the BLS field order.

    Ref: https://github.com/ethereum/EIPs/blob/master/EIPS/eip-2333.md#hkdf_mod_r
    """
    L = 48  # `ceil((3 * ceil(log2(r))) / 16)`, where `r` is the order of the BLS 12-381 curve
    salt = b"BLS-SIG-KEYGEN-SALT-"
    SK = 0
    while SK == 0:
        salt = SHA256(salt)
        okm = HKDF(
            salt=salt,
            IKM=IKM + b"\x00",  # add postfix `I2OSP(0, 1)`
            L=L,
            info=key_info + L.to_bytes(2, "big"),
        )
        SK = int.from_bytes(okm, byteorder="big") % bls_curve_order
    return SK


def PBKDF2(
    *, password: bytes, salt: bytes, dklen: int, c: int, prf: str
) -> bytes:
    if "sha" not in prf:
        raise ValueError(f"String 'sha' is not in `prf`({prf})")
    if "sha256" in prf and c < 2**18:
        """
        Verify the number of rounds of SHA256-PBKDF2. SHA512 not checked as use in BIP39
        does not require, and therefore doesn't use, safe parameters (c=2048).

        Ref: https://github.com/bitcoin/bips/blob/master/bip-0039.mediawiki#from-mnemonic-to-seed
        """
        raise ValueError("The PBKDF2 parameters chosen are not secure.")
    _hash = _sha256 if "sha256" in prf else _sha512
    res = _PBKDF2(
        password=password,
        salt=salt,
        dkLen=dklen,
        count=c,
        hmac_hash_module=_hash,
    )  # type: ignore
    # PyCryptodome can return Tuple[bytes]
    return res if isinstance(res, bytes) else res[0]


def scrypt(
    *, password: str, salt: str, n: int, r: int, p: int, dklen: int
) -> bytes:
    if n * r * p < 2**20:  # 128 MB memory usage
        raise ValueError("The Scrypt parameters chosen are not secure.")
    if n >= 2 ** (128 * r / 8):
        raise ValueError(
            "The given `n` should be less than `2**(128 * r / 8)`."
            f"\tGot `n={n}`, r={r}, 2**(128 * r / 8)={2**(128 * r / 8)}"
        )
    res = _scrypt(password=password, salt=salt, key_len=dklen, N=n, r=r, p=p)
    # PyCryptodome can return Tuple[bytes]
    return res if isinstance(res, bytes) else res[0]



from .abi import get_contract_abi
from .token import get_token_name
from .solidity import to_bytes, to_bytes32, to_string
from .crypto import keccak256, solidity_keccak256
from .id import generate_id, get_key
from .wrappers import http_request, multiple_attempt
from .bls import validate_deposit_data_file



from dotenv import dotenv_values
import sys
import os
import pytest
import inspect

sys.path.append(os.getcwd())

try:
    from geodefi import Geode
    from geodefi.classes.id import Id
    from geodefi.exceptions import MaxAttemptException
except ImportError:
    raise


env = dotenv_values(".env")


G = Geode(exec_api=env["EXECUTION_API"], cons_key=env["CONSENSUS_KEY"])


def test_init():
    id = Id(
        G.w3,
        G.network,
        G.Portal,
        id=50016835115526216130031110555486827201953559012021267556883950029143900999178,
    )

    assert id.w3 == G.w3
    assert id.network == G.network
    assert id.portal == G.Portal
    assert (
        id.ID
        == 50016835115526216130031110555486827201953559012021267556883950029143900999178
    )


def test_invalid_arguments():
    # Test case 3: Test with invalid arguments
    # with pytest.raises(TypeError):
    # Trying to initialize the class without providing parameters
    #    id = Id()

    # with pytest.raises(ValueError):
    # Trying to initialize the class with different parameter types

    #    id = Id(1, 2, 3, id="AAAAAA")

    pass


###################################################################
# Init is costly so below test will use this id instance.
###################################################################


id_portal = Id(
    G.w3,
    G.network,
    G.Portal,
    id=50016835115526216130031110555486827201953559012021267556883950029143900999178,
)
id_operator = Id(
    G.w3,
    G.network,
    G.Portal,
    id=50016835115526216130031110555486827201953559012021267556883950029143900999178,
)


def test_str():

    assert (
        str(id_portal)
        == "50016835115526216130031110555486827201953559012021267556883950029143900999178"
    )


def test_readUint():

    # test private variables
    assert id_portal._readUint("initiated") == 0
    assert id_portal._readUint("surplus") == 0
    assert id_portal._readUint("secured") == 0

    # test private variables
    assert id_operator._readUint("released") == 0
    assert id_operator._readUint("totalProposedValidators") == 0
    assert id_operator._readUint("priorFee") == 0


def test_invalid_readUint():

    # test incorrect values
    with pytest.raises(MaxAttemptException):
        id_portal._readUint(1231231)

    with pytest.raises(MaxAttemptException):
        id_portal._readUint(list("maintainer"))


"""
def test_readBytes():

    ## Only use-case of readBytes for now.
    assert id_portal._readBytes(
        "withdrawalCredential") == b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xc8.\xd5\xecW\x16s\xe6\xb1\x8cK\t,\x9c\xbcJ\xe8lxn'
    assert id_portal._readBytes(
        "withdrawalCredential", isHex= True) == '0x010000000000000000000000c82ed5ec571673e6b18c4b092c9cbc4ae86c786e'


def test_readBytesArray():

    # Only use-case of readBytes for now.
    assert id_portal._readBytesArray(
        "validators", index=0) == b'\x93&\xf6\xc0\x7f\x8a\xbd\x08.\xf8+\x19\'\x9c\xbb\xa7ak\x03\x95\xfb\x94}P\xcd-_\xef0=\xd6\x13\xab\xe3\x10\x87\x07zg\xfa\xa4w\xc0c\x1c\xc7"\x8d'

    # Index of empty storage
    assert id_portal._readBytesArray(
        "validators", index=1) == b''


def test_invalid_readBytesArray():

    # Negative index
    with pytest.raises(MaxAttemptException):
        id_portal._readBytesArray(
            "validators", index=-1)




def test_readAddress():

    assert id_portal._readAddress("whitelist") == 0
    assert id_portal._readAddress("withdrawalContract") == 0
    assert id_portal._readAddress("liquidityPool") == 0

"""


def test_NAME():

    assert isinstance(id_portal.NAME, str)  # == "Ice Bear's Pool"



from dotenv import dotenv_values
import sys
import os
import pytest
import inspect
import re

from geodefi.utils.merkle import StandartMerkleTree
from geodefi import Geode
from geodefi.classes.operator import Operator


sys.path.append(os.getcwd())


env = dotenv_values(".env")


G = Geode(exec_api=env["EXECUTION_API"], cons_key=env["CONSENSUS_KEY"])
PORTAL = G.Portal

OPERATOR_LIST = [
    114391297015478800753082638170652680401082080549997516459063441314156612391510,
    106440257855610140438147021303249733137151692982873052796064618175630395030085,
]


def test_init():

    myOperator = PORTAL.operator(OPERATOR_LIST[0])
    assert isinstance(myOperator, Operator)

    # == 114391297015478800753082638170652680401082080549997516459063441314156612391510
    assert isinstance(myOperator.ID, int)
    assert myOperator.TYPE == 4
    assert myOperator.network == 5

    assert myOperator.w3 == G.w3

    myOperator2 = PORTAL.operator(
        "114391297015478800753082638170652680401082080549997516459063441314156612391510"
    )
    assert str(myOperator.ID) == myOperator2.ID


myOperator = PORTAL.operator(OPERATOR_LIST[0])


def test_initiated():
    assert myOperator.initiated == 1687433352


def test_totalProposedValidators():
    assert myOperator.totalProposedValidators == 0


def test_totalActiveValidators():
    assert myOperator.totalActiveValidators == 0


def test_feeSwitch():
    assert myOperator.feeSwitch == 0


def test_priorFee():
    assert myOperator.priorFee == 0


def test_fee():
    assert myOperator.fee == 600000000


def test_periodSwitch():
    assert myOperator.periodSwitch == 0


def test_priorPeriod():
    assert myOperator.priorPeriod == 0


def test_validatorPeriod():
    assert myOperator.validatorPeriod == 15552000


def test_wallet():
    assert myOperator.wallet == 0


def test_released():
    assert myOperator.released == 0



from web3 import Web3
from Crypto.Hash import keccak as _keccak

from typing import List, Any


def keccak256(x: bytes) -> bytes:
    # Calculate Keccak-256 hash of the input bytes

    return _keccak.new(data=x, digest_bits=256).digest()


def solidity_keccak256(types: List[str], params: List[Any]):
    # Calculate Keccak-256 hash of the given solidity types and parameters
    return Web3.solidity_keccak(abi_types=types, values=params)
